#!/usr/bin/php -d memory_limit=128M
<?

/**  This is simple PHP code for PowerDNS PIPE backend. It provides basic functionality as speaking to PowerDNS via its PIPE Protocol.
        The lookups are delegated to special object. Abstract super class of this object is provided in the begining of this file.
	Author: Radek Hladik r.hladik@cybersales.cz
	You can use and/or modifiy and/or redistribute this code as long as you let me know about all modifications.
*/

$protocol_version="UNKNOWN";	//Variable for storing actual PIPE protocol version
$backend_name="PHP Backend Module ver 0.9";	//Default name and version of the backend. This should be overriden by object later
$statfile="/tmp/pdns-tinydns.stat"; //File for periodical stats
$counters=Array("queries"=>0,"malformed"=>0,"positive"=>0,"negative"=>0,"mintime"=>-1,"maxtime"=>-1,"sumtime"=>0,"valid"=>0); //Statistical counters
/** Abstract super class for lookup object
*/
class backend {
	/** Initialize the object. I.e. load data from external storage */
	function initialize()   
	{
	}
	
	/** Process the query - $data contains parsed query params as array
		$qname=$data[0];
		$qclass=$data[1];
		$qtype=$data[2];
		$id=$data[3];
		$ip=$data[4];
		$localip=$data[5];
		$localip need not to be specified - it depends on protocol used in communication. See http://doc.powerdns.com/backends-detail.html#PIPEBACKEND for more details
		
		This function should return records found in format specified in http://doc.powerdns.com/backends-detail.html#PIPEBACKEND without the ending end. It can additionaly set error using get_error method. The error will be appended as LOG to the output.
		This function must not return final END keyword. Example of returned data:
		DATA            jupiter.heaven.af.mil   IN      A       2560    1       192.168.1.2
	*/
	function process_query($data)
	{
		
	}
	/** Should list the whole domain. $data should contain record ID from previous SOA query
	Example:
	Q	ds9a.nl		IN	SOA	-1	213.244.168.210
	DATA	ds9a.nl		IN	SOA	86400	1 ahu.ds9a.nl ...
	END
	AXFR		1
	DATA	ds9a.nl		IN	SOA	86400	1 ahu.ds9a.nl ...
	DATA	ds9a.nl		IN	NS	86400	1 ns1.ds9a.nl
	DATA	ds9a.nl		IN	NS	86400	1 ns2.ds9a.nl
	DATA	ns1.ds9a.nl	IN	A	86400	1 213.244.168.210
	DATA	ns2.ds9a.nl	IN	A	86400	1 63.123.33.135
	END
	
	Notice the 1 returned in SOA as record ID. This is later reused for AXFR.
	*/
	function process_axfr($data)
	{
		
	}
	
	/** Return name and version of the backend */
	function get_name()
	{
		return ("Abstract backend");
	}
	
	/** Must return boolean. True if the object is ready to answer query (i.e. all data are loaded) or false if object is not ready (i.e. object was not able to read configuration file) */
	function is_ready()
	{
		return (false);
	}
	
	/** Return last error if any or "" if no error was raised. Reset error to no error state */
	function get_error()
	{
		return ("");
	}
	
	/** Return count of domains/zones loaded */
	function get_domain_count()
	{
		return(0);
	}
	
	/** Return count of records loaded */
	function get_record_count()
	{
		return(0);
	}	
	/** Return string (with \n after each line) with backend object statistics */
	function get_backend_stats()
	{
		return("");
	}
}

/** Get current microtime in float (source www.php.net) */
function microtime_float()
{
    list($usec, $sec) = explode(" ", microtime());
    return ((float)$usec + (float)$sec);
}


$backend=null;
require ("/root/pdns/tinydns.php");  // Load backend. This included file must also create an instance of backend object into $backend variable

if (is_object($backend)) $backend_name=$backend->get_name();


function process_query($data)  //Parse query, do some checking, pass query to object if available and output response+error
{	global $backend,$counters;
	$response="";
	$tmparr=split("\t",$data);
	$counters["queries"]++;
	if (count($tmparr)<5) 
	{
		$response="LOG\tMalformed query received ".(count($tmparr))."\n";
		$counters["malformed"]++;
	}
	else
	{
		$counters["valid"]++;
		$tmparr[5]=((isset($tmparr[5]))?$tmparr[5]:"");
		$time_start = microtime_float();
		$tmpresponse=$backend->process_query($tmparr);
		$time_end = microtime_float();
		$time = $time_end - $time_start;
		if (($time > $counters["maxtime"]) || ($counters["maxtime"]==-1)) $counters["maxtime"]=$time;
		if (($time < $counters["mintime"]) || ($counters["mintime"]==-1)) $counters["mintime"]=$time;
		$counters["sumtime"]+=$time;
		$response.=$tmpresponse;
		if ($tmpresponse=="") $counters["negative"]++; else $counters["positive"]++;
		if (($error=$backend->get_error())!="") $response.="LOG\t$error\n";
		//$response.="LOG\t".$tmparr[0]." ".$tmparr[2]."\n";
	}
	$response.="END\n";
	return ($response);
}

function process_axfr($data)  //Basic AXFR processing. Pass to Backend object
{
	global $backend,$counters;
	$response="";
	$tmpresponse=$backend->process_axfr($data);
	$response.=$tmpresponse;
	if (($error=$backend->get_error())!="") $response.="LOG\t$error\n";
	$response.="END\n";
	return ($response);

}

function generate_stats()  //Generate stats string
{
	global $counters,$backend;
	$result="";
	if (is_object($backend))
	{
	$result.="Total zones		 ".$backend->get_domain_count()."\n";
	$result.="Total records		 ".$backend->get_record_count()."\n";
	}
	$memory=number_format(memory_get_usage()/(1024*1024),2,".","");
	$result.="Memory used		 ".$memory."MB\n";
	$result.="Total queries		 ".$counters["queries"]."\n";
	$result.="Malformed queries	 ".$counters["malformed"]."\n";
	$result.="Valid queries		 ".$counters["valid"]."\n";
	$result.="Min processing time	".$counters["mintime"]."s\n";
	$result.="Max processing time	 ".$counters["maxtime"]."s\n";
	$result.="Avg processing time 	 ".(($counters["valid"]>0)?(1.0*$counters["sumtime"]/$counters["valid"]."s"):"N/A")."\n";
	if (is_object($backend)) $result.=$backend->get_backend_stats();
	return ($result);
}



function process_command($command,$data)  //process command
{	global $protocol_version,$backend_name,$backend,$counters,$statfile;
	$response="";
	if ($command=="HELO")  //HELO command
	{
		if (($data=="1") || ($data=="2"))  //Protocol version 1
		{
			$protocol_version=$data;
			$time_start = microtime_float();
			if (is_object($backend)) $backend->initialize();    //Initialize object
			$time_end = microtime_float();
			$time = $time_end - $time_start;
			if (is_object($backend)) $domcount=$backend->get_domain_count(); else $domcount=0;  //Get count of zones loaded by object
			$loadtime=number_format  ( $time, 3 , ".","" );
			$memory=number_format(memory_get_usage()/(1024*1024),2,".","");
			$response="OK\t".$backend_name." (PID ".getmypid()."):  ".$domcount." zones with ".$backend->get_record_count()." records loaded in ".$loadtime."s using ".$memory."MB\n";  //Output OK and Banner
			
			
		}
		else
		{
			$response="FAIL\n";   //Unknow protocol version
		}
		
	}
	else if ($command=="PING")  //Response to PING command
	{
		$response="END\n";
	}
	else if ($command=="STAT")  //Response to STAT command - experimental, not defined by PowerDNS
	{
		$response=generate_stats();
		$response.="END\n";
	}
	else if ($command=="AXFR")  //Response to AXFR
	{
		if (is_object($backend)) //Verify if backend object exists
		{
			if ($backend->is_ready())   //Verify if backend object is ready to process queries
			{
				$response=process_axfr($data);
		
			}
			else
			{
				$response="LOG\tPHP Backend not ready";
				if (($berror=$backend->get_error())!="") $response .= trim($berror);
				$response .= "\n";
				$response.="END\n";
			}
		} 
		else
		{
			$response="LOG\tNo PHP Backend available\nEND\n";
		}
		
	}
	else if ($command=="Q")  //Response to Query command
	{
		if (is_object($backend)) //Verify if backend object exists
		{
			if ($backend->is_ready())   //Verify if backend object is ready to process queries
			{
				$response=process_query($data);
				$fp=fopen ($statfile,"w");fputs($fp,generate_stats());fclose($fp); //Generate stats to stat file - temporary solution, just for  performance estimation
			}
			else
			{
				$response="LOG\tPHP Backend not ready";
				if (($berror=$backend->get_error())!="") $response .= trim($berror);
				$response .= "\n";
				$response.="END\n";
			}
		} 
		else
		{
			$response="LOG\tNo PHP Backend available\nEND\n";
		}
		
	}
	
	return ($response);
}

function parse_input($line) //Parse line of output. Check for some strange inputs (empty line etc) and call process command if command received
{
	$result=Array("data"=>"","error"=>"unknown error");
	if ($line=="") 
		{ 
			$result["error"]=="Empty line on input"; 
			return ($result) ; 
		}
	$lineparts=split("\t",$line,2);
	$command=$lineparts[0];
	if ($command=="") 
		{ 
			$result["error"]=="Empty command on input"; 
			return ($result) ; 
		}
		$result["data"]=process_command($lineparts[0],((count($lineparts)==1)?"":$lineparts[1]));
	$result["error"]="";	
	return ($result);
}


// main routine
  $forcequit=false;
  while (!$forcequit)  //indefinite loop for reading input
  {
	  $line=fgets(STDIN);
   	  $result=parse_input(trim($line));
	  if ($result["error"]=="") 
	  {
		  echo ($result["data"]);
		  
	  } else
	  {
		  echo ($result["error"]);
	  }
  }


?>
