<?
      
//cache array=cca 100Bytes per record
//record array=cca 800Bytes per record (for ANY search speed optimization is record stored twice)
/** Convert tai64 formated time to UNIX timestamp */
function tai64_to_timestamp($tai64_number){

/***  remove @ sign ***/
$tai64_number = str_replace('@', '', $tai64_number);

/*** convert to unix timestamp and hexdec ***/
return intval(hexdec($tai64_number));
}

/**  This is the class for tinydns zonefile loading, parsing and serving to backend.php (simple PowerDNS pipe backend wrapper)
        
	Author: Radek Hladik r.hladik@cybersales.cz
	You can use and/or modifiy and/or redistribute this code as long as you let me know about all modifications.
*/
class backend_tinydns extends backend
{
	public $filename="/root/pdns/data";  //Zone file
	public $lastchange=0; // Last zonefile change
	public $ready=false; // Is backend ready to process queries?
	public $error=""; // Last error
	public $data=null; //Parsed data
	public $default_ttl=2560;  //Default ttl for records (
	public $default_refresh=16384; //Default refresh
	public $default_retry=2048; // Default retry
	public $default_expire=1048576; //Default expire
	public $default_default_ttl=86400; //Default default_ttl - All default values shlould be the same as in tinydns - however this one is little bit tricky. Feel free to tune them  
	public $location_cache=Array();  //Storage for location cache
	public $location_cache_limit=100000; //Records max in cache
	public $location_cache_count=0; //Records max in cache
	public $record_id=1; //Record ID counter
	public $record_count=0; //Records count
	public $rrtypes=Array(
				      1=>"A",
				      2=>"NS",
				      5=>"CNAME",
				      6=>"SOA",
				      9=>"MR",
				      12=>"PTR",
				      13=>"HINFO",
				      16=>"MX",
				      16=>"TXT",
				      17=>"RP",
				      18=>"AFSDB",
				      24=>"SIG",
				      25=>"KEY",
				      28=>"AAAA",
				      29=>"LOC",
				      33=>"SRV",
				      38=>"A6",
				      35=>"NAPTR",
				      99=>"SPF",
				      252=>"AXFR",
				      255=>"ANY",
				      256=>"URL",
				      257=>"MBOXFW",
				      258=>"CURL"
					);   //Types for converting generic RRs to its appropriate PowerDNS names
	
	/** Initilaize the plugin - load data */
	function initialize()  
	{
		$this->load_data();
	}
	
	/** Calculate TTL for record.
	$recordttl - TTL specified in zone line
	$name - record name - for zone defaultTTL lookup 
	*/
	function get_ttl($recordttl,$name="")
	{
		$ttl=-1;
		if (($recordttl>0) || ($recordttl==="0")) $ttl=$recordttl;  //Set Record specific TTL 
		if (($name != "")  && ($ttl==-1))  //If no record specific TTL set and we have name, try to get TTL from domain soa
		{
			$domain="";
			if (($dotpos=strpos($name,"."))!=false) $domain=substr($name,$dotpos+1); 
			if (isset($this->data["domains"][$name])) 
			{
				$ttl=$this->data["domains"][$name]["DEFAULT_TTL"];
			} else if (isset($this->data["domains"][$domain]))
			{
				$ttl=$this->data["domains"][$domain]["DEFAULT_TTL"];
			}
			
		}
		if ($ttl==-1) $ttl=$this->default_ttl;  //In all other cases get default TTL
		return ($ttl);
	}
	
	/** Add parsed record to storage */
	function add_record($type,$name,$data)
	{
		if (!isset($this->data[$type])) $this->data[$type]=Array();  //If this record is first of its kind
		$data["TYPE"]=$type;
		if ($data["TIMESTAMP"] != "" ) {$data["TIMESTAMP"]=tai64_to_timestamp($data["TIMESTAMP"]); }  //Convert timestamp
		$data["ID"]=(++$this->record_id);  
		$this->data[$type][$name][]=$data; //Add to storage for specific record type
		$this->data["ANY"][$name][]=$data; //Add to storage for ANY (all ) record types
		$this->record_count++;
	}
	/** Parse input file and load zone data */
	function load_data()
	{
		if (! file_exists($this->filename)) {$this->error="Filename $filename not found"; return; }
		if (! is_readable($this->filename)) {$this->error="No read access for $filename"; return; }
		$statarr=stat($this->filename);
		try
		{
			$data=file($this->filename);
		}
		catch (Exception $e) {
			$this->error="Error reading $filename (".$e->getMessage().")"; 
			return;
		}
		$this->lastchange=$statarr["mtime"]; //remember last modification time for further zonedata reloading
		
		$this->ready=false;
		$this->data=Array("ANY"=>Array(),"A"=>Array(),"MX"=>Array(),"NS"=>Array(),"SOA"=>Array(),"PTR"=>Array(),"domains"=>Array()); //Prepare storage for most common RRs
		$this->record_count=0;
		foreach ($data as $line) //Parse line by line
		{
			$line=trim($line);
			$tag=substr($line,0,1); //Get first character - record tag which specifies RR type
			if (($tag=="#") || ($tag=="")) //comment
			{
			}
			else if ($tag=="%")  //location definition
			{  
				$dataarr=explode(":",substr($line,1).":");
				$this->data["locations"][$dataarr[0]]=$dataarr[1];
			}
			else if (($tag=="=") || ($tag=="+"))  //A RR  with or without automatic reverse PTR (= with, + without)
			{	//fqdn:ip:ttl:timestamp:lo
				$dataarr=explode(":",substr($line,1).":::::");
				if (($dataarr[0]!="") && ($dataarr[1]!=""))	
				{
					$this->add_record("A",$dataarr[0],Array("VALUE"=>$dataarr[1],"TTL"=>$this->get_ttl($dataarr[2],$dataarr[0]),"TIMESTAMP"=>$dataarr[3],"LOCATION"=>$dataarr[4]));
					//a PTR (``pointer'') record showing fqdn as the name of d.c.b.a.in-addr.arpa if ip is a.b.c.d.
					$iparr=explode('.',$dataarr["1"]);
					if (($tag=="=") && (count($iparr)==4))  //Try to reverse IP
					{
						$reverseip="";
						$ipok=true; foreach ($iparr as $ippart) if (($ippart>0) && ($ippart<256)) {$reverseip=$ippart.".".$reverseip; } else {$ipok=false;}
						if ($ipok) $this->add_record("PTR",$reverseip."in-addr.arpa",Array("VALUE"=>$dataarr[0],"TTL"=>$this->get_ttl($dataarr[2],$dataarr[0]),"TIMESTAMP"=>$dataarr[3],"LOCATION"=>$dataarr[4]));
					}
				}
		
			}
			else if ($tag=="Z")  //SOA - "plain SOA"
			{	//Zfqdn:mname:rname:ser:ref:ret:exp:min:ttl:timestamp:lo
				$dataarr=explode(":",substr($line,1).":::::::::::::");
				if (($dataarr[0]!="") && ($dataarr[1]!=""))	
				{
					$hostmaster=$dataarr[2];if ($hostmaster=="") $hostmaster="hostmaster.".$dataarr[0];  //Calculate implicit hostmaster name if not specified in line record
					if (strpos($hostmaster,'.')>-1) $hostmaster[strpos($hostmaster,'.')]='@'; //replace first dot with @
					$serial=$dataarr[3];
					if (! ($serial>0)) $serial=$this->lastchange; //If serial is not an number use default - timestamp of zone file
					$default_ttl=(($dataarr[7]>0)?$dataarr[7]:$this->default_default_ttl); 
					$soavalue=$dataarr[1]." ".$hostmaster." ".$serial." ".(($dataarr[4]>0)?$dataarr[4]:$this->default_refresh)." ".(($dataarr[5]>0)?$dataarr[5]:$this->default_retry)." ".(($dataarr[6]>0)?$dataarr[6]:$this->default_expire)." ".$default_ttl;
					$this->add_record("SOA",$dataarr[0],Array("VALUE"=>$soavalue,"TTL"=>$this->get_ttl($dataarr[8],$dataarr[0]),"TIMESTAMP"=>$dataarr[9],"LOCATION"=>$dataarr[10]));
					$this->data["domains"][$dataarr[0]]=Array("DEFAULT_TTL"=>$default_ttl); //Store default ttl for domain
				}
		
			}
			else if (($tag==".") || ($tag=="&"))  //SOA with NS and optional A record
			{
				//.fqdn:ip:x:ttl:timestamp:lo
				$dataarr=explode(":",substr($line,1)."::::::");
				if (($dataarr[0]!=""))
				{
					list($name,$ip,$value,$ttl,$timestamp,$location)=$dataarr;
					$ip=$dataarr[1];if ($ip=="") $ip="127.0.0.1"; //For records like .localhost
					$ttl=$this->get_ttl($ttl,$name);
					
					if (strpos($value,".")===false) $ns=$value.".ns.".$name; else $ns=$value; //If x contains a dot then tinydns-data will use x as the server name rather than x.ns.fqdn
					if ($value=="") $ns="ns.".$name;
					
					$this->add_record("NS",$name,Array("VALUE"=>$ns,"TTL"=>$ttl,"TIMESTAMP"=>$timestamp,"LOCATION"=>$location));
						
					if ($ip != "") $this->add_record("A",$ns,Array("VALUE"=>$ip,"TTL"=>$ttl,"TIMESTAMP"=>$timestamp,"LOCATION"=>$location));
					
					if (($tag==".") && (! isset($this->data["SOA"][$name])))
					{
						$serial=$this->lastchange;
						$soavalue=$ns." hostmaster@$name ".$serial." ".($this->default_refresh)." ".($this->default_retry)." ".($this->default_expire)." ".$this->default_default_ttl;
						$this->add_record("SOA",$name,Array("VALUE"=>$soavalue,"TTL"=>$ttl,"TIMESTAMP"=>$timestamp,"LOCATION"=>$location));
						$this->data["domains"][$dataarr[0]]=Array("DEFAULT_TTL"=>$this->default_default_ttl);
					}
					
						
				}
			}
			else if (($tag=="@"))  //MX
			{
				// @fqdn:ip:x:dist:ttl:timestamp:lo
				$dataarr=explode(":",substr($line,1).":::::::::");
				list($name,$ip,$value,$prio,$ttl,$timestamp,$location)=$dataarr;
				$ttl=$this->get_ttl($ttl,$name);
					
				if (strpos($value,".")===false) $mx=$value.".mx.".$name; else $mx=$value; //If x contains a dot then tinydns-data will use x as the server name rather than x.mx.fqdn
					
				$this->add_record("MX",$name,Array("VALUE"=>$prio."\t".$mx,"TTL"=>$ttl,"TIMESTAMP"=>$timestamp,"LOCATION"=>$location));
				
				
			}
			else if (($tag=="'"))  //TXT - tag is '
			{
				//'fqdn:s:ttl:timestamp:lo
				$dataarr=explode(":",substr($line,1).":::::::");
				list($name,$value,$ttl,$timestamp,$location)=$dataarr;
				$value=stripcslashes($value);
				$ttl=$this->get_ttl($ttl,$name);
				$this->add_record("TXT",$name,Array("VALUE"=>$value,"TTL"=>$ttl,"TIMESTAMP"=>$timestamp,"LOCATION"=>$location));
				
			}
			else if (($tag=="^"))  //PTR
			{
				//^fqdn:p:ttl:timestamp:lo
				$dataarr=explode(":",substr($line,1)."::::::::");
				list($name,$value,$ttl,$timestamp,$location)=$dataarr;
				$ttl=$this->get_ttl($ttl,$name);
				$this->add_record("PTR",$name,Array("VALUE"=>$value,"TTL"=>$ttl,"TIMESTAMP"=>$timestamp,"LOCATION"=>$location));
				
			}
			else if (($tag=="C"))  //CNAME
			{
				//Cfqdn:p:ttl:timestamp:lo
				$dataarr=explode(":",substr($line,1)."::::::::");
				list($name,$value,$ttl,$timestamp,$location)=$dataarr;
				$ttl=$this->get_ttl($ttl,$name);	
				$this->add_record("CNAME",$name,Array("VALUE"=>$value,"TTL"=>$ttl,"TIMESTAMP"=>$timestamp,"LOCATION"=>$location));
				
			}
			else if (($tag==":"))  //Generic record - Generic record for fqdn. tinydns-data creates a record of type n for fqdn showing rdata. n must be an integer between 1 and 65535; it must not be 2 (NS), 5 (CNAME), 6 (SOA), 12 (PTR), 15 (MX), or 252 (AXFR).
			{
				//:fqdn:n:rdata:ttl:timestamp:lo
				$dataarr=explode(":",substr($line,1).":::::::");
				list($name,$type,$value,$ttl,$timestamp,$location)=$dataarr;
				$ttl=$this->get_ttl($ttl,$name);
				if (isset($this->rrtypes[$type]))
				{
					$pdnstype=$this->rrtypes[$type];
					//$value=stripcslashes($value);
					$this->add_record($pdnstype,$name,Array("VALUE"=>$value,"TTL"=>$ttl,"TIMESTAMP"=>$timestamp,"LOCATION"=>$location));
				}
			}
			
			else
			{
				
				//Unknown tag
				echo "Unknown tag *".$tag."* \n";
			}
		}
		$this->ready=true; //Finally we are ready to process queries
	}
	
	/** Process query */
	function process_query($data)
	{
		$result="";
		if (! $this->ready) {$this->error="Backend not ready for query"; return; }
		if (! is_array($this->data)) {$this->error="Backend data not ready for query"; return; }
		if ( count($data)<5) {$this->error="Not enough query arguments"; return; }
		if ( count($data)>6) {$this->error="Too many query arguments"; return; }
		
		clearstatcache(); /* Check whether zonefile has not changed */
		$statarr=stat($this->filename);
		if ($this->lastchange != $statarr["mtime"])
		{
			$result .= "LOG\tData file changed - reloading zones\n"; /* If so - reload it */
			$this->load_data();
		}
		
		
		list ($qname,$qclass,$qtype,$id,$ip,$localip)=$data; //Load query parametrs 
		if ( $qclass!="IN") {$this->error="Unknown address class $qclass"; return; } 
		
		if (( ! isset($this->data[$qtype])) || (! is_array($this->data[$qtype]))) return("");  //Return nothig if we do not have any record of requested type
		if (( ! isset($this->data[$qtype][$qname])) || (! is_array($this->data[$qtype][$qname]))) return(""); //Return nothing if do not have requested record
		
		$userlocation=false;
		foreach ($this->data[$qtype][$qname] as $record)  //Parse all records found
		{
			$recordactive=true;  //Record will be returned by default
			if ($record["TIMESTAMP"]!="")  //If timestamp is specified decide records fate based on current time
			{
				$ttlcheck=$this->check_timestamp($record["TIMESTAMP"],$record["TTL"]);
				if ($ttlcheck != true) $recordactive=false; 
			}
			if ($record["LOCATION"] != "") //If location is specified decide records fate based on request IP
			{
				if ($userlocation===false) $userlocation=$this->lookup_location($ip);
				if ($userlocation!=$record["LOCATION"]) $recordactive=false;
			}
			if ($recordactive) $result.="DATA	\t$qname\t$qclass\t".$record["TYPE"]."\t".$record["TTL"]."\t".$record["ID"]."\t".$record["VALUE"]."\n"; //Output record if it made it 
		}
		return ($result);
		
	}
	/** Basic AXFR processing  - probably not very efficient but usable */
	function process_axfr($data)
	{
		$record=$data;  //As AXFR argument we have record of SOA obtained by previous SOA query
		$foundsoa=false;
		$response="";
		
		clearstatcache(); //check for zonefile change
		$statarr=stat($this->filename);
		if ($this->lastchange != $statarr["mtime"])
		{
			$response .= "LOG\tData file changed - reloading zones\n";
			$this->load_data();
		}
		
		
		foreach ($this->data["SOA"] as $key=>$soas)  //search all SOAs for given record ID
			foreach ($soas as $soa)
			{
				if ($soa["ID"]==$record) $foundsoa=$key;  //We've found it
			}
		if ($foundsoa==false) { $this->error="Zone with ID $record not found"; return; }  //Return error if zone has not been found
		
		$qclass="IN";
		foreach ($this->data["ANY"] as $key=>$data)  //Return all records we have for this domain
			{
				$domainpart=substr($key,strpos($key,'.')+1);  //Strip hostname part of name  and return all records which match (either by domain or by whole name, i.e.  for example.com return records www.example.com and example.com)
				if (($domainpart==$foundsoa) || ($key==$foundsoa)) foreach ($data as $record) $response.="DATA	\t$key\t$qclass\t".$record["TYPE"]."\t".$record["TTL"]."\t".$record["ID"]."\t".$record["VALUE"]."\n";
			}
		return ($response);
	}
	
	/** Get name and version of this module */
	function get_name()
	{
		return ("PHP TinyDNS Backend ver 0.9");
	}
	
	/** Is this module ready to process queries? */
	function is_ready()
	{
		return ($this->ready);
	}
	
	/** Get and reset last error */
	function get_error()
	{
		$error=$this->error; $this->error="";
		return ($error);
	}
	
	/** Get count of loaded zones - statistical counter */
	function get_domain_count()
	{
		return(count($this->data["domains"]));
	}
	
	/** Get count of loaded records - statistical counter */
	function get_record_count()
	{
		return($this->record_count);
	}
	
	/** Check timestamp against current time - If ttl is nonzero (or omitted), the timestamp is a starting time for the information in the line; the line will be ignored before that time. If ttl is zero, the timestamp is an ending time (``time to die'') for the information in the line */
	function check_timestamp($ts,$ttl)
	{
		if ($ttl=="0") 	return ($ts>time()); else return ($ts<=time()); 
	}
	
	/** Lookup requester location */
	function lookup_location($ip)
	{
		if (! isset($this->location_cache[$ip]))  //if record is not in cache - look ip up and store it in the cache
		{
			$userlocation="";
			$userloclen=-1;
			foreach ($this->data["locations"] as $locname=>$location) //go through all defined locations and find the longest that matches 
			{
				if ((substr($ip,0,$loclen=strlen($location))==($location)) && ($loclen>$userloclen)) { $userlocation=$locname; $userloclen=$loclen;} 
			}
			//echo "User $ip is in location $userlocation \n";
			$this->location_cache_count++;
			if ($this->location_cache_count==$this->location_cache_limit) { $this->location_cache=Array(); $this->location_cache_count=0; } //flush cache if too big
			$this->location_cache[$ip]=$userlocation; //store in the cacge
		}
		return ($this->location_cache[$ip]); //return from cache
		
	}
	/** Provide some basic stats about module */
	function get_backend_stats()
	{
		$result="Location cache		".$this->location_cache_count." items\n";
		return ($result);
	}
	
	
}



$backend=new backend_tinydns(); // Create instance to global variable


?>
